local BOM = BuffomatAddon ---@type BomAddon

---@shape BomActionCastModule
local actionCastModule = BomModuleManager.actionCastModule

local buffomatModule = BomModuleManager.buffomatModule
local taskModule = BomModuleManager.taskModule
local partyModule = BomModuleManager.partyModule
local allBuffsModule = BomModuleManager.allBuffsModule

---@class BomTaskActionCast: BomTaskAction Casts a spell with a power (mana) cost on a target
---@field buffDef BomBuffDefinition|nil
---@field spellLink string|nil
---@field spellId number|nil
---@field manaCost number
---@field temporaryDownrank boolean Pick previous rank for certain spells, like Flametongue 10
---@field target BomUnit
local actionCastClass = {}
actionCastClass.__index = actionCastClass

---@param cost number Resource cost (mana cost)
---@param spellId number Spell id to capture
---@param link string Spell link for a picture
---@param targetUnit BomUnit player to benefit from the spell
---@param buffDef BomBuffDefinition the spell to be added
---@param temporaryDownrank boolean Pick previous rank for certain spells, like Flametongue 10
---@return BomTaskActionCast
function actionCastModule:New(cost, spellId, link, targetUnit, buffDef, temporaryDownrank)
  local a = --[[---@type BomTaskActionCast]] {}
  setmetatable(a, actionCastClass)
  a.temporaryDownrank = temporaryDownrank
  a.manaCost = cost
  a.spellId = spellId
  a.spellLink = link
  a.buffDef = buffDef
  a.target = targetUnit
  return a
end

function actionCastClass:CanCast()
  local cdtest = GetSpellCooldown(self.spellId) or 0
  if cdtest ~= 0 then
    BOM.checkCooldown = self.spellId
    --BomC_ListTab_Button:Disable()
    return taskModule.CAN_CAST_ON_CD
  else
    --BomC_ListTab_Button:Enable()
  end
  BOM.castFailedBuff = self.buffDef
  BOM.castFailedBuffTarget = self.target

  if self.buffDef
          and (--[[---@not nil]] self.buffDef).type ~= "resurrection"
          and self.target.isDead then
    -- Cannot cast buffs on deads, only resurrections
    return taskModule.CAN_CAST_IS_DEAD
  end

  if self.manaCost > partyModule.playerMana then
    return taskModule.CAN_CAST_OOM
  end

  return taskModule.CAN_CAST_OK
end

---@param task BomTask
function actionCastClass:GetButtonText(task)
  return (self.spellLink or "?") .. " " .. task.extraText
end

function actionCastClass:GetCancelFormMacroLine()
  local _, playerClass, _ = UnitClass("player")

  if playerClass == "DRUID" then
    return "/cancelform [nocombat, noform:5]"
  end

  return "/cancelform [nocombat]"
end

---@param m BomMacro
function actionCastClass:UpdateMacro(m)
  --Downgrade-Check
  local buffDef = allBuffsModule.buffFromSpellIdLookup[--[[---@not nil]] self.spellId]
  local rank = ""

  if buffDef == nil then
    BOM:Debug("Update macro: buffDef is nil for spellid=" .. tostring(self.spellId))
    return
  end

  if buffomatModule.shared.UseRank
          or (self.target and (--[[---@not nil]] self.target).unitId == "target")
  then
    local level = UnitLevel((--[[---@not nil]] self.target).unitId)

    if buffDef and level ~= nil and level > 0 then
      local spellChoices

      if buffDef.singleFamily
              and tContains(buffDef.singleFamily, self.spellId) then
        spellChoices = buffDef.singleFamily
      elseif buffDef.groupFamily
              and tContains(--[[---@not nil]] buffDef.groupFamily, self.spellId) then
        spellChoices = buffDef.groupFamily
      end

      if spellChoices then
        local newSpellId

        for i, id in ipairs(spellChoices) do
          if buffomatModule.shared.SpellGreaterEqualThan[id] == nil
                  or buffomatModule.shared.SpellGreaterEqualThan[id] < level then
            newSpellId = id
          else
            break
          end
          if id == self.spellId then
            break
          end
        end
        self.spellId = newSpellId or self.spellId
      end
    end -- if spell and level
  end

  if self.temporaryDownrank then
    self.spellId = self.buffDef:GetDownRank(self.spellId)
    rank = GetSpellSubtext(--[[---@not nil]] self.spellId) or ""

    if rank ~= "" then
      rank = "(" .. rank .. ")"
    end
    --BOM:Debug("rank=" .. rank .. " spellid=" .. self.spellId)
  end

  BOM.castFailedSpellId = self.spellId
  local name = GetSpellInfo(--[[---@not nil]] self.spellId)
  if name == nil then
    BOM:Debug("Update macro: Bad spellid=" .. tostring(self.spellId))
    return
  end

  if tContains(allBuffsModule.cancelForm, self.spellId) then
    table.insert(m.lines, self:GetCancelFormMacroLine())
  end
  table.insert(m.lines, "/bom _checkforerror")

  table.insert(m.lines, "/cast [@"
          .. (--[[---@not nil]] self.target).unitId
          .. ",nocombat]" .. name .. rank)
end
